#include <sys/time.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>

static int N=100;
#define NB_STEPS 50

/* uncomment this line to activate the debug mode */
//#define DEBUG 1
#ifdef DEBUG
#define DEBUG_PRINTF printf
#else
#define DEBUG_PRINTF //printf
#endif

double **cur_step;
double **next_step;

int comm_rank;
int comm_size;

/* initialize matrixes */
void init() {
  int i, j;

  cur_step = malloc(sizeof(double*) * N);
  next_step = malloc(sizeof(double*) * N);

  srand(time(NULL));
  for(i=0; i<N; i++) {
    cur_step[i] = calloc(N, sizeof(double));
    next_step[i] = calloc(N, sizeof(double));
  }

  int nb_hot_spots= rand()%N;
  printf("Generating %d hot spots\n", nb_hot_spots);
  for(i=0; i<nb_hot_spots; i++) {
    int posx = rand()%N;
    int posy = rand()%N;
    cur_step[posx][posy] += N*comm_size*10000;
    printf("%d,%d : %lf\n", posx, posy, cur_step[posx][posy]);
  }
}

/* dump the matrix in f */
void print_matrix(FILE* f,double** matrix) {
  int i, j;
  for(i=1; i<N-1; i++) {
    for(j=1; j<N-1; j++) {
      fprintf(f, " %.2f  ", matrix[i][j]);
    }
    fprintf(f, "\n");
  }
}

/* dump all the submatrixes in a file */
void print_all(FILE* f) {
  int i, j;
  
  /* first, print the local matrix */
  print_matrix(f, cur_step);

  /* TODO:
   *  - each MPI rank sends its matrix to rank 0
   *  - rank 0 receives all the matrices and write them to f
   */
}


void compute() {
  int i, j;
  MPI_Request r1, r2;

  for(i=1; i< N-1; i++) {
    for(j=1; j< N-1; j++) {
      next_step[i][j] = (cur_step[i-1][j] + cur_step[i+1][j] +
			 cur_step[i][j-1] + cur_step[i][j+1] +
			 cur_step[i][j]) / 5;
    }
  }
  
  DEBUG_PRINTF("rank #%d: compute done\n", comm_rank);

  /* TODO: each process exchange its frontiers with its neighbours */


  /* swap buffers */
  double ** tmp = cur_step;
  cur_step = next_step;
  next_step = tmp;
}


int main(int argc, char**argv) {

  char* output_file="result.dat";
  int dump=1;
  MPI_Init(&argc, &argv);

  MPI_Comm_size(MPI_COMM_WORLD, &comm_size);
  MPI_Comm_rank(MPI_COMM_WORLD, &comm_rank);

  if(argc>1){
    N = atoi(argv[1]);
  }

  struct timeval t1, t2;
  if(!comm_rank)
    printf("Initialization (problem size: %d)\n", N);
  init();

  MPI_Barrier(MPI_COMM_WORLD);
  if(!comm_rank)
    printf("Start computing (%d steps)\n", NB_STEPS);

  gettimeofday(&t1, NULL);
  int i;

  for(i=0; i< NB_STEPS; i++) {
    if(!comm_rank)
      printf("STEP %d...\n", i);
    compute();
  }

  gettimeofday(&t2, NULL);
  MPI_Barrier(MPI_COMM_WORLD);

  if(comm_rank==0) {
    double total_time = ((t2.tv_sec-t1.tv_sec)*1e6 + (t2.tv_usec - t1.tv_usec))/1e6;
    printf("%d steps in %lf sec (%lf sec/step)\n",
	   NB_STEPS, total_time, total_time/NB_STEPS);
  }

  if(dump){
    if(!comm_rank)
      printf("dumping the result data in %s\n", output_file);
    FILE *f = fopen(output_file, "w");
    print_all(f);
  }

  MPI_Finalize();
  return 0;
}
